@@ -1,9 +1,16 @@
 """Discovery module for TP-Link Smart Home devices."""
 import asyncio
+import binascii
 import logging
 import socket
 from typing import Awaitable, Callable, Dict, Optional, Type, cast
 
+# When support for cpython older than 3.11 is dropped
+# async_timeout can be replaced with asyncio.timeout
+from async_timeout import timeout as asyncio_timeout
+
+from kasa.credentials import Credentials
+from kasa.exceptions import UnsupportedDeviceException
 from kasa.json import dumps as json_dumps
 from kasa.json import loads as json_loads
 from kasa.protocol import TPLinkSmartHomeProtocol
@@ -36,13 +43,26 @@ class _DiscoverProtocol(asyncio.DatagramProtocol):
         target: str = "255.255.255.255",
         discovery_packets: int = 3,
         interface: Optional[str] = None,
-    ):
+        on_unsupported: Optional[Callable[[Dict], Awaitable[None]]] = None,
+        port: Optional[int] = None,
+        discovered_event: Optional[asyncio.Event] = None,
+        credentials: Optional[Credentials] = None,
+        timeout: Optional[int] = None,
+    ) -> None:
         self.transport = None
         self.discovery_packets = discovery_packets
         self.interface = interface
         self.on_discovered = on_discovered
-        self.target = (target, Discover.DISCOVERY_PORT)
+        self.discovery_port = port or Discover.DISCOVERY_PORT
+        self.target = (target, self.discovery_port)
+        self.target_2 = (target, Discover.DISCOVERY_PORT_2)
         self.discovered_devices = {}
+        self.unsupported_devices: Dict = {}
+        self.invalid_device_exceptions: Dict = {}
+        self.on_unsupported = on_unsupported
+        self.discovered_event = discovered_event
+        self.credentials = credentials
+        self.timeout = timeout
 
     def connection_made(self, transport) -> None:
         """Set socket options for broadcasting."""
@@ -67,25 +87,48 @@ class _DiscoverProtocol(asyncio.DatagramProtocol):
         req = json_dumps(Discover.DISCOVERY_QUERY)
         _LOGGER.debug("[DISCOVERY] %s >> %s", self.target, Discover.DISCOVERY_QUERY)
         encrypted_req = TPLinkSmartHomeProtocol.encrypt(req)
-        for i in range(self.discovery_packets):
+        for _i in range(self.discovery_packets):
             self.transport.sendto(encrypted_req[4:], self.target)  # type: ignore
+            self.transport.sendto(Discover.DISCOVERY_QUERY_2, self.target_2)  # type: ignore
 
     def datagram_received(self, data, addr) -> None:
         """Handle discovery responses."""
         ip, port = addr
-        if ip in self.discovered_devices:
+        if (
+            ip in self.discovered_devices
+            or ip in self.unsupported_devices
+            or ip in self.invalid_device_exceptions
+        ):
             return
 
-        info = json_loads(TPLinkSmartHomeProtocol.decrypt(data))
-        _LOGGER.debug("[DISCOVERY] %s << %s", ip, info)
+        if port == self.discovery_port:
+            info = json_loads(TPLinkSmartHomeProtocol.decrypt(data))
+            _LOGGER.debug("[DISCOVERY] %s << %s", ip, info)
+
+        elif port == Discover.DISCOVERY_PORT_2:
+            info = json_loads(data[16:])
+            self.unsupported_devices[ip] = info
+            if self.on_unsupported is not None:
+                asyncio.ensure_future(self.on_unsupported(info))
+            _LOGGER.debug("[DISCOVERY] Unsupported device found at %s << %s", ip, info)
+            if self.discovered_event is not None:
+                self.discovered_event.set()
+            return
 
         try:
             device_class = Discover._get_device_class(info)
         except SmartDeviceException as ex:
-            _LOGGER.debug("Unable to find device type from %s: %s", info, ex)
+            _LOGGER.debug(
+                "[DISCOVERY] Unable to find device type from %s: %s", info, ex
+            )
+            self.invalid_device_exceptions[ip] = ex
+            if self.discovered_event is not None:
+                self.discovered_event.set()
             return
 
-        device = device_class(ip)
+        device = device_class(
+            ip, port=port, credentials=self.credentials, timeout=self.timeout
+        )
         device.update_from_discover_info(info)
 
         self.discovered_devices[ip] = device
@@ -93,6 +136,9 @@ class _DiscoverProtocol(asyncio.DatagramProtocol):
         if self.on_discovered is not None:
             asyncio.ensure_future(self.on_discovered(device))
 
+        if self.discovered_event is not None:
+            self.discovered_event.set()
+
     def error_received(self, ex):
         """Handle asyncio.Protocol errors."""
         _LOGGER.error("Got error: %s", ex)
@@ -123,7 +169,8 @@ class Discover:
         >>> [dev.alias for dev in found_devices]
         ['TP-LINK_Power Strip_CF69']
 
-        Discovery can also be targeted to a specific broadcast address instead of the 255.255.255.255:
+        Discovery can also be targeted to a specific broadcast address instead of
+        the default 255.255.255.255:
 
         >>> asyncio.run(Discover.discover(target="192.168.8.255"))
 
@@ -142,6 +189,9 @@ class Discover:
         "system": {"get_sysinfo": None},
     }
 
+    DISCOVERY_PORT_2 = 20002
+    DISCOVERY_QUERY_2 = binascii.unhexlify("020000010000000000000000463cb5d3")
+
     @staticmethod
     async def discover(
         *,
@@ -150,20 +200,27 @@ class Discover:
         timeout=5,
         discovery_packets=3,
         interface=None,
+        on_unsupported=None,
+        credentials=None,
     ) -> DeviceDict:
         """Discover supported devices.
 
         Sends discovery message to 255.255.255.255:9999 in order
         to detect available supported devices in the local network,
         and waits for given timeout for answers from devices.
-        If you have multiple interfaces, you can use target parameter to specify the network for discovery.
+        If you have multiple interfaces,
+        you can use *target* parameter to specify the network for discovery.
 
-        If given, `on_discovered` coroutine will get awaited with a :class:`SmartDevice`-derived object as parameter.
+        If given, `on_discovered` coroutine will get awaited with
+        a :class:`SmartDevice`-derived object as parameter.
 
-        The results of the discovery are returned as a dict of :class:`SmartDevice`-derived objects keyed with IP addresses.
-        The devices are already initialized and all but emeter-related properties can be accessed directly.
+        The results of the discovery are returned as a dict of
+        :class:`SmartDevice`-derived objects keyed with IP addresses.
+        The devices are already initialized and all but emeter-related properties
+        can be accessed directly.
 
-        :param target: The target address where to send the broadcast discovery queries if multi-homing (e.g. 192.168.xxx.255).
+        :param target: The target address where to send the broadcast discovery
+         queries if multi-homing (e.g. 192.168.xxx.255).
         :param on_discovered: coroutine to execute on discovery
         :param timeout: How long to wait for responses, defaults to 5
         :param discovery_packets: Number of discovery packets to broadcast
@@ -177,8 +234,11 @@ class Discover:
                 on_discovered=on_discovered,
                 discovery_packets=discovery_packets,
                 interface=interface,
+                on_unsupported=on_unsupported,
+                credentials=credentials,
+                timeout=timeout,
             ),
-            local_addr=("0.0.0.0", 0),
+            local_addr=("0.0.0.0", 0),  # noqa: S104
         )
         protocol = cast(_DiscoverProtocol, protocol)
 
@@ -193,21 +253,98 @@ class Discover:
         return protocol.discovered_devices
 
     @staticmethod
-    async def discover_single(host: str) -> SmartDevice:
+    async def discover_single(
+        host: str,
+        *,
+        port: Optional[int] = None,
+        timeout=5,
+        credentials: Optional[Credentials] = None,
+    ) -> SmartDevice:
         """Discover a single device by the given IP address.
 
+        It is generally preferred to avoid :func:`discover_single()` and
+        use :func:`connect_single()` instead as it should perform better when
+        the WiFi network is congested or the device is not responding
+        to discovery requests.
+
         :param host: Hostname of device to query
         :rtype: SmartDevice
         :return: Object for querying/controlling found device.
         """
-        protocol = TPLinkSmartHomeProtocol(host)
+        loop = asyncio.get_event_loop()
+        event = asyncio.Event()
+        transport, protocol = await loop.create_datagram_endpoint(
+            lambda: _DiscoverProtocol(
+                target=host,
+                port=port,
+                discovered_event=event,
+                credentials=credentials,
+                timeout=timeout,
+            ),
+            local_addr=("0.0.0.0", 0),  # noqa: S104
+        )
+        protocol = cast(_DiscoverProtocol, protocol)
 
-        info = await protocol.query(Discover.DISCOVERY_QUERY)
+        try:
+            _LOGGER.debug("Waiting a total of %s seconds for responses...", timeout)
 
-        device_class = Discover._get_device_class(info)
-        dev = device_class(host)
-        await dev.update()
+            async with asyncio_timeout(timeout):
+                await event.wait()
+        except asyncio.TimeoutError as ex:
+            raise SmartDeviceException(
+                f"Timed out getting discovery response for {host}"
+            ) from ex
+        finally:
+            transport.close()
 
+        if host in protocol.discovered_devices:
+            dev = protocol.discovered_devices[host]
+            await dev.update()
+            return dev
+        elif host in protocol.unsupported_devices:
+            raise UnsupportedDeviceException(
+                f"Unsupported device {host}: {protocol.unsupported_devices[host]}"
+            )
+        elif host in protocol.invalid_device_exceptions:
+            raise protocol.invalid_device_exceptions[host]
+        else:
+            raise SmartDeviceException(f"Unable to get discovery response for {host}")
+
+    @staticmethod
+    async def connect_single(
+        host: str,
+        *,
+        port: Optional[int] = None,
+        timeout=5,
+        credentials: Optional[Credentials] = None,
+    ) -> SmartDevice:
+        """Connect to a single device by the given IP address.
+
+        This method avoids the UDP based discovery process and
+        will connect directly to the device to query its type.
+
+        It is generally preferred to avoid :func:`discover_single()` and
+        use this function instead as it should perform better when
+        the WiFi network is congested or the device is not responding
+        to discovery requests.
+
+        The device type is discovered by querying the device.
+
+        :param host: Hostname of device to query
+        :rtype: SmartDevice
+        :return: Object for querying/controlling found device.
+        """
+        unknown_dev = SmartDevice(
+            host=host, port=port, credentials=credentials, timeout=timeout
+        )
+        await unknown_dev.update()
+        device_class = Discover._get_device_class(unknown_dev.internal_state)
+        dev = device_class(
+            host=host, port=port, credentials=credentials, timeout=timeout
+        )
+        # Reuse the connection from the unknown device
+        # so we don't have to reconnect
+        dev.protocol = unknown_dev.protocol
         return dev
 
     @staticmethod
