@@ -37,23 +37,17 @@ class TPLinkSmartHomeProtocol:
     DEFAULT_TIMEOUT = 5
     BLOCK_SIZE = 4
 
-    def __init__(self, host: str, *, port: Optional[int] = None) -> None:
+    def __init__(
+        self, host: str, *, port: Optional[int] = None, timeout: Optional[int] = None
+    ) -> None:
         """Create a protocol object."""
         self.host = host
         self.port = port or TPLinkSmartHomeProtocol.DEFAULT_PORT
         self.reader: Optional[asyncio.StreamReader] = None
         self.writer: Optional[asyncio.StreamWriter] = None
-        self.query_lock: Optional[asyncio.Lock] = None
+        self.query_lock = asyncio.Lock()
         self.loop: Optional[asyncio.AbstractEventLoop] = None
-
-    def _detect_event_loop_change(self) -> None:
-        """Check if this object has been reused betwen event loops."""
-        loop = asyncio.get_running_loop()
-        if not self.loop:
-            self.loop = loop
-        elif self.loop != loop:
-            _LOGGER.warning("Detected protocol reuse between different event loop")
-            self._reset()
+        self.timeout = timeout or TPLinkSmartHomeProtocol.DEFAULT_TIMEOUT
 
     async def query(self, request: Union[str, Dict], retry_count: int = 3) -> Dict:
         """Request information from a TP-Link SmartHome Device.
@@ -64,19 +58,12 @@ class TPLinkSmartHomeProtocol:
         :param retry_count: how many retries to do in case of failure
         :return: response dict
         """
-        self._detect_event_loop_change()
-
-        if not self.query_lock:
-            self.query_lock = asyncio.Lock()
-
         if isinstance(request, dict):
             request = json_dumps(request)
-            assert isinstance(request, str)
-
-        timeout = TPLinkSmartHomeProtocol.DEFAULT_TIMEOUT
+            assert isinstance(request, str)  # noqa: S101
 
         async with self.query_lock:
-            return await self._query(request, retry_count, timeout)
+            return await self._query(request, retry_count, self.timeout)
 
     async def _connect(self, timeout: int) -> None:
         """Try to connect or reconnect to the device."""
@@ -90,8 +77,8 @@ class TPLinkSmartHomeProtocol:
 
     async def _execute_query(self, request: str) -> Dict:
         """Execute a query on the device and wait for the response."""
-        assert self.writer is not None
-        assert self.reader is not None
+        assert self.writer is not None  # noqa: S101
+        assert self.reader is not None  # noqa: S101
         debug_log = _LOGGER.isEnabledFor(logging.DEBUG)
 
         if debug_log:
@@ -121,7 +108,7 @@ class TPLinkSmartHomeProtocol:
 
     def _reset(self) -> None:
         """Clear any varibles that should not survive between loops."""
-        self.reader = self.writer = self.loop = self.query_lock = None
+        self.reader = self.writer = None
 
     async def _query(self, request: str, retry_count: int, timeout: int) -> Dict:
         """Try to query a device."""
@@ -129,11 +116,11 @@ class TPLinkSmartHomeProtocol:
         # Most of the time we will already be connected if the device is online
         # and the connect call will do nothing and return right away
         #
-        # However, if we get an unrecoverable error (_NO_RETRY_ERRORS and ConnectionRefusedError)
-        # we do not want to keep trying since many connection open/close operations
-        # in the same time frame can block the event loop. This is especially
-        # import when there are multiple tplink devices being polled.
-        #
+        # However, if we get an unrecoverable error (_NO_RETRY_ERRORS and
+        # ConnectionRefusedError) we do not want to keep trying since many
+        # connection open/close operations in the same time frame can block
+        # the event loop.
+        # This is especially import when there are multiple tplink devices being polled.
         for retry in range(retry_count + 1):
             try:
                 await self._connect(timeout)
@@ -141,26 +128,28 @@ class TPLinkSmartHomeProtocol:
                 await self.close()
                 raise SmartDeviceException(
                     f"Unable to connect to the device: {self.host}:{self.port}: {ex}"
-                )
+                ) from ex
             except OSError as ex:
                 await self.close()
                 if ex.errno in _NO_RETRY_ERRORS or retry >= retry_count:
                     raise SmartDeviceException(
-                        f"Unable to connect to the device: {self.host}:{self.port}: {ex}"
-                    )
+                        f"Unable to connect to the device:"
+                        f" {self.host}:{self.port}: {ex}"
+                    ) from ex
                 continue
             except Exception as ex:
                 await self.close()
                 if retry >= retry_count:
                     _LOGGER.debug("Giving up on %s after %s retries", self.host, retry)
                     raise SmartDeviceException(
-                        f"Unable to connect to the device: {self.host}:{self.port}: {ex}"
-                    )
+                        f"Unable to connect to the device:"
+                        f" {self.host}:{self.port}: {ex}"
+                    ) from ex
                 continue
 
             try:
-                assert self.reader is not None
-                assert self.writer is not None
+                assert self.reader is not None  # noqa: S101
+                assert self.writer is not None  # noqa: S101
                 async with asyncio_timeout(timeout):
                     return await self._execute_query(request)
             except Exception as ex:
